(defun key (s) (first s))          ; key at root
(defun dat (s) (second s))         ; data at root
(defun lft (s) (third s))          ; left subtree
(defun rgt (s) (fourth s))         ; right subtree
(defun emptyp (s) (not (consp s))) ; empty tree?
(defun size (s)                    ; number of keys
  (if (emptyp s)
      0                                     ; {sz0}
      (+ 1 (size (lft s)) (size (rgt s))))) ; {sz1}
(defun iskeyp (k)
   (natp k))
; DEFINITIONS REVISED FOR FAST HEIGHT OPERATION
(defun ht (s) ; extract height of tree s
  (if (emptyp s)
      0
      (fifth s)))
(defun mktr (k d lf rt) ; make tree from key, data, and subtrees
    (list k d lf rt (+ 1 (max (ht lf) (ht rt)))))
(defun treep (s) ; search tree?
  (or (emptyp s)
      (and (= (len s) 5) (iskeyp (key s))
           (treep (lft s)) (treep (rgt s))
           (= (ht s) (+ 1 (max (ht(lft s)) (ht(rgt s))))))))
(defun balp (s) ; tree s is balanced?
  (or (emptyp s)
      (and (<= (abs (- (ht (lft s)) (ht (rgt s)))) 1)
           (balp (lft s)) (balp (rgt s)))))
; END OF REVISED DEFINITIONS
(defun keyp (k s)                ; key k occurs in s?
  (and (iskeyp k) (treep s) (not (emptyp s))
       (or (= k (key s)) (keyp k (lft s)) (keyp k (rgt s)))))
(defun hook (x a s) ; put a new key x with data a into tree s
  (if (emptyp s)    ; preserve order, but not necessarily balance ; WARNING! typo in book
      (mktr x a nil nil)                     ; {hook0}
      (let* ((k  (key s)) (d (dat s))
             (lf (lft s)) (rt (rgt s)))
        (if (< x k)
            (mktr k d (hook x a lf) rt)      ; {hook<}
            (if (> x k)
                (mktr k d lf (hook x a rt))  ; {hook>}
                (mktr x a lf rt))))))        ; {hook=}
(defun zig (s) ; rotate clockwise
  (let* ((k  (key s)) (d (dat s))
         (j  (key (lft s))) (c  (dat (lft s)))
         (jL (lft (lft s))) (jR (rgt (lft s)))
         (kR (rgt s)))
    (mktr j c jL (mktr k d jR kR))))
(defun zag (s) ; rotate counterclockwise
  (let* ((k  (key s)) (d (dat s))
         (kL (lft s))
         (i (key (rgt s))) (b (dat (rgt s)))
         (iL (lft (rgt s))) (iR (rgt (rgt s))))
    (mktr i b (mktr k d kL iL) iR)))
; ROTATIONS REVISED FOR FAST HEIGHT OPERATION
(defun rot+ (s)           ; rotate clockwise if too tall on left
  (let* ((k  (key s)) (d  (dat s)) ; rot+ assumes s is not empty
         (kL (lft s)) (kR (rgt s)))
    (if (> (ht kL) (+ (ht kR) 1))           ; unbalanced?
        (if (< (ht (lft kL)) (ht (rgt kL))) ; inside lft?
            (zig(mktr k d (zag kL) kR))     ; dbl rotate
            (zig s))                        ; sngl rotate
        s)))                                ; no rotate
(defun rot- (s)    ; rotate counterclockwise if too tall on right
  (let* ((k (key s)) (d (dat s))    ; rot- assumes s is not empty
         (kL (lft s)) (kR (rgt s)))
    (if (> (ht kR) (+ (ht kL) 1))           ; unbalanced?
        (if (> (ht (lft kR)) (ht (rgt kR))) ; inside rgt?
            (zig(mktr k d kL (zag kR)))     ; dbl rotate
            (zig s))                        ; sngl rotate
        s)))                                ; no rotate
; END REVISED ROTATIONS
(defun ins (x a s)
  (if (emptyp s)
      (mktr x a nil nil)                          ; one-node tree
      (let* ((k  (key s)) (d  (dat s))
             (kL (lft s)) (kR (rgt s)))
        (if (< k x)
            (rot+ (mktr k d (ins x a kL) kR))     ; insert left
            (if (> k x)
                (rot- (mktr k d kL (ins x a kR))) ; insert right
                (mktr x a kL kR))))))             ; new root data